#include "socket.h"
#include <errno.h>
#include <string.h>
#include <string>

#ifdef WIN32
#pragma comment(lib, "ws2_32.lib")
#define EPROTOTYPE WSAEPROTOTYPE
#else
#include <netdb.h>
#endif

/*******************************************************************
 * socket class member functions
 *
 *******************************************************************/

socket::socket(int domain, int type)
{
    id = ::socket(domain, type, 0);
    if ((int)id<0) {
        throw strerror_s(errno);
    }
}

socket::~socket()
{
    close();
}

void socket::bind(struct sockaddr *adr, int len)
{
    int result = ::bind(id, adr, len);
    if (result<0) {
        throw strerror_s(errno);
    }
}

void socket::listen()
{
    int result = ::listen(id, 5);
    if (result<0) {
        throw strerror_s(errno);
    }
}

void socket::connect(struct sockaddr *adr, int len)
{
    int result = ::connect(id, adr, len);
    if (result<0) {
        throw strerror_s(errno);
    }
}

void socket::accept(socket *sock, struct sockaddr *adr, int *adrlen)
{
    sock->id = ::accept(id, adr, (socklen_t*)adrlen);
    if (sock->id < 0) {
        throw strerror_s(errno);
    }
}

int socket::send(const char *buffer, int len)
{
    int result = ::send(id, buffer, len, 0);
    if (result<0) {
        throw strerror_s(errno);
    }
    return result;
}

template<class T> int socket::send(const T& buffer)
{
    size_t size = (buffer.end() - buffer.begin()) * sizeof(typename T::value_type);
    int result = ::send(id, buffer.data(), int(size), 0);
    if (result<0) {
        throw strerror_s(errno);
    }
    return result;
}

int socket::sendto(const char *buffer, int len, struct sockaddr *adr, int adrlen)
{
    int result = ::sendto(id, buffer, len, 0, adr, adrlen);
    if (result<0) {
        throw strerror_s(errno);
    }
    return result;
}

template<class T> int socket::sendto(const T& buffer, struct sockaddr *adr, int adrlen)
{
    size_t size = (buffer.end() - buffer.begin()) * sizeof(typename T::value_type);
    int result = ::sendto(id, buffer.data(), int(size), adr, adrlen);
    if (result<0) {
        throw strerror_s(errno);
    }
    return result;
}
    
int socket::recv(char *buffer, int len)
{
    int result = ::recv(id, buffer, len, 0);
    if (result<0) {
        throw strerror_s(errno);
    }
    return result;
}

int socket::recvfrom(char *buffer, int len, struct sockaddr *adr, int *adrlen)
{
    int result = ::recvfrom(id, buffer, len, 0, adr, (socklen_t*)adrlen);
    if (result<0) {
        throw strerror_s(errno);
    }
    return result;
}

void socket::close()
{
    ::shutdown(id, 2);
#ifndef WIN32
    ::close(id);
#endif    
}

string socket::strerror_s(int e)
{
    return string(strerror(e));
}



/*******************************************************************
 * ip_socket class member functions
 *
 *******************************************************************/
 
ip_socket::ip_socket(int type)
    : socket(AF_INET, type)
{
}

void ip_socket::bind(unsigned int port, unsigned int ipadr)
{
    struct sockaddr_in sadr;
    sadr.sin_family = AF_INET;
    sadr.sin_port = htons(port);
    sadr.sin_addr.s_addr = ipadr;
    socket::bind((sockaddr*)&sadr, sizeof(sockaddr_in));
}

void ip_socket::accept(ip_socket &sock, peer_info &pi)
{
    struct sockaddr_in sadr;
    int slen = sizeof(sockaddr_in);
    socket::accept(&sock, (sockaddr*)&sadr, &slen);
    if (sadr.sin_family != AF_INET) {
        throw strerror_s(EPROTOTYPE);
    }
    pi.adr = ipv4_adr(ntohl(sadr.sin_addr.s_addr));
    pi.port = ntohs(sadr.sin_port);
}

string ip_socket::gethostname(ipv4_adr adr)
{
    struct in_addr inadr;
    *(unsigned int*)&inadr = adr.get_n();
    struct hostent *h = ::gethostbyaddr((const char *)&inadr, sizeof(struct in_addr), AF_INET);
    if (h && h->h_name) {
        return string(h->h_name);
    } else {
        return "";
    }
}

void ip_socket::gethostadr(vector<ipv4_adr> &result, const string &name)
{
    result.clear();
    struct hostent *hp = ::gethostbyname(name.c_str());
    for (int i=0; hp->h_addr_list[i]; ++i) {
        result.push_back(ipv4_adr((in_addr*)&(hp->h_addr_list[i])));
    }
}

/*******************************************************************
 * tcp_socket class member functions
 *
 *******************************************************************/

tcp_socket::tcp_socket()
    : ip_socket(SOCK_STREAM)
{    
}

void tcp_socket::connect(const ipv4_adr &adr, unsigned int port)
{
    struct sockaddr_in sadr;
    sadr.sin_family = AF_INET;
    sadr.sin_port = htons(port);
    sadr.sin_addr.s_addr = htonl(adr.get());
    socket::connect((sockaddr*)&sadr, sizeof(sockaddr_in));
}

void tcp_socket::connect(const string &host, unsigned int port)
{
    struct hostent *hp = ::gethostbyname(host.c_str());
    for (int i=0; hp->h_addr_list[i]; ++i) {
        try {
            connect(ipv4_adr(inet_ntoa(*(struct in_addr*)(hp->h_addr_list[i]))), port);
            return;
        } catch (...) {
        }
    }
    throw strerror_s(errno);
}


/*******************************************************************
 * udp_socket class member functions
 *
 *******************************************************************/

udp_socket::udp_socket()
    : ip_socket(SOCK_DGRAM)
{
}

int udp_socket::sendto(const char *buffer, int len, const ipv4_adr &adr, unsigned int port)
{
    struct sockaddr_in sadr;
    sadr.sin_family = AF_INET;
    sadr.sin_port = htons(port);
    sadr.sin_addr.s_addr = htonl(adr.get());
    return socket::sendto(buffer, len, (sockaddr*)&sadr, sizeof(sockaddr));
}

template<class T> int udp_socket::sendto(const T& buffer, unsigned int ipadr, unsigned int port)
{
    struct sockaddr_in sadr;
    sadr.sin_family = AF_INET;
    sadr.sin_port = htons(port);
    sadr.sin_addr.s_addr = htonl(ipadr);
    return socket::sendto(buffer, (sockaddr*)&sadr, sizeof(sockaddr));
}

int udp_socket::recvfrom(char *buffer, int len, peer_info &pi)
{
    struct sockaddr_in sadr;
    int slen = sizeof(sockaddr_in);
    int result = socket::recvfrom(buffer, len, (sockaddr*)&sadr, &slen);
    if (sadr.sin_family != AF_INET) {
        throw strerror_s(EPROTOTYPE);
    }
    if (result<0) {
        throw strerror_s(errno);
    }
    pi.adr = ipv4_adr(ntohl(sadr.sin_addr.s_addr));
    pi.port = ntohs(sadr.sin_port);
    return result;
}

/*******************************************************************
 * forced instantiations
 *
 *******************************************************************/

template int socket::send<std::string>(const std::string& buffer);
